@@ -76,10 +76,11 @@ def load_config(N=None):
         raise RuntimeError("No configuration file found.")
     cf = ConfigParser()
     if N is not None:
+        if not os.path.isfile(N):
+            raise RuntimeError("%s not found." % (N))
         cf.read(N)
     else:
         cf.read(cfgfile)
-
     return cf
 
 
@@ -143,17 +144,21 @@ def main():
                              'by default, uses EPICS7 pvAccess.'
                              ' e.g. ca://MY:TEST:TRIG:PV or MY:TEST:TRIG:PV')
     parser.add_argument(
-        "--log-level",
-        dest="loglevel",
-        default="error",
-        help="log level. Accepts lower or uppercase values. Defaults to ERROR",
+        "--debug",
+        action='store_true',
+        help="Turn on debug messages"
     )
 
     args = parser.parse_args()
 
+    level = logging.WARNING
+
+    if args.debug:
+        level = logging.DEBUG
+        
     logging.basicConfig(
         stream=sys.stdout,
-        level=args.loglevel.upper(),
+        level=level,
         format="%(asctime)s %(levelname)-6s %(message)s",
         datefmt="%Y-%m-%d %H:%M:%S",
     )
@@ -167,7 +172,7 @@ def main():
     app = args.app
     if args.app is None:
         app = AppType(cfg["DEFAULT"]["APP"].lower())
-    
+        
     if app == AppType.IMAGE:
         from c2dataviewer import imagev
         # application from CLI overwrite the one in configuration file
